package crypto

import (
	"crypto/aes"
	"crypto/cipher"
	"crypto/rand"
	"io"
)

// CryptoAlgorithm is an interface that provides the methods required for
// encryption and decryption of data.
// Currently only AES is supported, but this could be extended to support other
// algorithms in the future.
type CryptoAlgorithm interface {
	Encrypt(plaintext []byte, key []byte) ([]byte, []byte, error)
	Decrypt(nonce, ciphertext []byte, key []byte) ([]byte, error)
}

// AES is an implementation of the CryptoAlgorithm interface.
// AES provides the methods required for performing symmetric key encryption
// and decryption using the AES algorithm.
//
// For this the aes package from the Go standard library is used.
type AES struct{}

// NewAES creates a new instance of a AES struct.
func NewAES() *AES {
	return &AES{}
}

// Encrypt encrypts the plaintext using a provided key.
// The key should have a length of 16, 24 or 32 bytes to select AES-128,
// AES-192 or AES-256.
// The method returns the nonce, the encrypted output and an error if something
// went wrong.
func (a *AES) Encrypt(plaintext []byte, key []byte) ([]byte, []byte, error) {
	// create a new cipher block from the key
	c, err := aes.NewCipher(key)
	if err != nil {
		return nil, nil, err
	}

	// create a new block cipher wrapped in GCM with default nonce (12
	// bytes) and tag size (16 bytes).
	gcm, err := cipher.NewGCM(c)
	if err != nil {
		return nil, nil, err
	}

	// generate a random nonce of nonce size (12 bytes)
	nonce := make([]byte, gcm.NonceSize())
	if _, err = io.ReadFull(rand.Reader, nonce); err != nil {
		return nil, nil, err
	}

	// Encrypt the plaintext using AES-GCM
	// Destination is set to nil, therefore seal only contains the
	// ciphertext with the tag appended.
	seal := gcm.Seal(nil, nonce, plaintext, nil)

	return nonce, seal, nil
}

// Decrypt decrypts the ciphertext using the provided key and nonce.
// The key should have a length of 16, 24 or 32 bytes to select AES-128,
// AES-192 or AES-256.
// The method returns the decrypted input.
func (a *AES) Decrypt(nonce, ciphertext []byte, key []byte) ([]byte, error) {
	// create a new cipher block from the key
	c, err := aes.NewCipher(key)
	if err != nil {
		return nil, err
	}

	// create a new block cipher wrapped in GCM with default nonce (12
	// bytes) and tag size (16 bytes).
	gcm, err := cipher.NewGCM(c)
	if err != nil {
		return nil, err
	}

	// Decrypt the ciphertext using AES-GCM
	return gcm.Open(nil, nonce, ciphertext, nil)
}
