package peers

import (
	"context"
	"errors"
	"fmt"
	"sync"
	"time"

	pbIC "code.fbi.h-da.de/danet/quant/goKMS/api/gen/proto/go/kmsintercom"
	"code.fbi.h-da.de/danet/quant/goKMS/kms/event"
	"code.fbi.h-da.de/danet/quant/goKMS/kms/store"
	"code.fbi.h-da.de/danet/quant/goKMS/kms/util"
	"code.fbi.h-da.de/danet/quant/quantumlayer"
	"github.com/google/uuid"
	log "github.com/sirupsen/logrus"
)

type DanetQuantumModule struct {
	QlID  uuid.UUID
	kmsId string
	// QuantumElementLink *quantumlayer.QuantumlayerEmuPRNG // contains information about the quantum links
	// key stores of unchopped bulk keys go here
	addr             string
	RawBulkKeysMutex sync.Mutex
	RawBulkKeys      map[int64]*quantumlayer.QuantumLayerBulkKey
	keyStore         *store.KmsKeyStore // the keys used between two peers.
	kmsClient        *GRPCClient
	kmsEventBus      *event.EventBus
	kmsTcpSocketStr  string
}

func NewDanetQuantumModule(kmsUDPAddr string, kmsId string) *DanetQuantumModule {
	return &DanetQuantumModule{
		QlID:            uuid.New(),
		kmsId:           kmsId,
		addr:            kmsUDPAddr,
		RawBulkKeys:     make(map[int64]*quantumlayer.QuantumLayerBulkKey),
		keyStore:        store.NewKmsKeyStore(256),
		kmsClient:       nil,
		kmsEventBus:     nil,
		kmsTcpSocketStr: "",
	}
}

func (qm *DanetQuantumModule) ID() uuid.UUID {
	return qm.QlID
}

func (qm *DanetQuantumModule) Initialize() error {
	return nil
}

func (qm *DanetQuantumModule) SetKmsPeerInformation(kmsClient *GRPCClient, kmsEventBus *event.EventBus, kmsTcpSocketStr string) error {
	qm.kmsClient = kmsClient
	qm.kmsEventBus = kmsEventBus
	qm.kmsTcpSocketStr = kmsTcpSocketStr
	return nil
}

func (qm *DanetQuantumModule) Address() string {
	return qm.addr
}

func (qm *DanetQuantumModule) Sync() error {
	rawBulkKeyIds := util.KeysOfMap(qm.RawBulkKeys)
	log.Info("Found the following bulk key ids for usage: ", rawBulkKeyIds)

	ctx, cancel := context.WithTimeout(context.Background(), time.Second)
	defer cancel()
	initialPeerSetupResponse, err := qm.kmsClient.SyncQkdBulk(ctx, &pbIC.SyncQkdBulkRequest{
		Timestamp: time.Now().Unix(),
		KmsId:     qm.kmsId,
		BulkId:    rawBulkKeyIds,
	})
	if err != nil {
		return err
	}

	bulkKey, ok := qm.RawBulkKeys[initialPeerSetupResponse.BulkId]
	if !ok {
		return fmt.Errorf("could not find raw bulk key with id: %d", initialPeerSetupResponse.BulkId)
	}

	// NOTE: Currently it is  assumed that only 256 bit keys are necessary.
	// This process could be improved by letting the peer partners initially
	// discuss about the key length.
	qm.keyStore = store.NewKmsKeyStore(256)

	keyIds, keyData, err := qm.KeyChopper(bulkKey, []string{})
	if err != nil {
		return err
	}

	_, err = qm.kmsClient.SyncKeyIdsForBulk(ctx, &pbIC.SyncKeyIdsForBulkRequest{
		Timestamp: time.Now().Unix(),
		KmsId:     qm.kmsId,
		BulkId:    initialPeerSetupResponse.BulkId,
		KeyId:     keyIds,
	})
	if err != nil {
		return err
	}

	for keyId, key := range keyData {
		qm.keyStore.AddKey(keyId, key)
	}

	qm.RawBulkKeysMutex.Lock()
	delete(qm.RawBulkKeys, initialPeerSetupResponse.BulkId)
	qm.RawBulkKeysMutex.Unlock()

	// Send notification about change
	if qm.kmsEventBus != nil {
		go func() {
			err := qm.kmsEventBus.Publish(event.NewPeerEvent(qm.kmsTcpSocketStr))
			if err != nil {
				log.Error(err)
			}
		}()
	}

	return nil
}

func (qm *DanetQuantumModule) KeyStore() *store.KmsKeyStore {
	return qm.keyStore
}

func (qm *DanetQuantumModule) SetKeyStore(newKeyStore *store.KmsKeyStore) {
	qm.keyStore = newKeyStore
}

func (qm *DanetQuantumModule) KmsGrpcClient() *GRPCClient {
	return qm.kmsClient
}

func (qm *DanetQuantumModule) SetKmsGrpcClient(peer *GRPCClient) {
	qm.kmsClient = peer
}

// Takes a bulk of keys and chops them in chopFactor keys each
// Any remainder is discarded
// If keyIds is empty, uuids are generated.
func (qm *DanetQuantumModule) KeyChopper(bulkKey *quantumlayer.QuantumLayerBulkKey, keyIds []string) ([]string, map[uuid.UUID][]byte, error) {
	if qm.keyStore.KeySingleSize == 0 {
		return nil, nil, errors.New("KeyChopper: no  keySingleSize set")
	}

	if bulkKey.BulkKeyLength != len(bulkKey.BulkKey) {
		if (bulkKey.BulkKeyLength) != (len(bulkKey.BulkKey) * 8) {
			log.Errorf("Length missmatch, even when converting 'bulkKey.BulkKey' to bytes: %d, %d", bulkKey.BulkKeyLength, len(bulkKey.BulkKey))
			return nil, nil, errors.New("bulkKey length mismatch")
		}
	}

	initialKeyIdsLen := len(keyIds)

	// Let's chop!
	keyData := make(map[uuid.UUID][]byte)
	chopFactor := qm.keyStore.KeySingleSize >> 3
	key := bulkKey.BulkKey
	counter := 0
	for len(key) > int(chopFactor) {
		var keyId uuid.UUID
		var err error
		if initialKeyIdsLen == 0 {
			keyId = uuid.New()
			keyIds = append(keyIds, keyId.String())
		} else {
			keyId, err = uuid.Parse(keyIds[counter])
			if err != nil {
				return nil, nil, fmt.Errorf("the provided ID: %s can not be parsed as UUID", keyIds[counter])
			}
			counter++
		}
		tmpkey := key[:chopFactor]
		keyData[keyId] = tmpkey

		// shorten the key storage
		key = key[chopFactor:]
	}
	return keyIds, keyData, nil
}
