package peers

import (
	"context"
	"fmt"
	"net/http"
	"net/url"
	"time"

	etsi14ClientGenerated "code.fbi.h-da.de/danet/quant/etsi014/go/rest/etsi/client"
	pbIC "code.fbi.h-da.de/danet/quant/goKMS/api/gen/proto/go/kmsintercom"
	"code.fbi.h-da.de/danet/quant/goKMS/config"
	etsi14ClientImpl "code.fbi.h-da.de/danet/quant/goKMS/kms/etsi/etsi14/client"
	"code.fbi.h-da.de/danet/quant/goKMS/kms/event"
	"code.fbi.h-da.de/danet/quant/goKMS/kms/store"
	kmstls "code.fbi.h-da.de/danet/quant/goKMS/kms/tls"
	"github.com/google/uuid"
	log "github.com/sirupsen/logrus"
)

type ETSI014HTTPQuantumModule struct {
	id          uuid.UUID
	kmsId       string
	addr        string
	keyStore    *store.KmsKeyStore
	kmsClient   *GRPCClient
	client      *etsi14ClientImpl.ClientImpl
	SlaveSAEID  string
	MasterSAEID string
	master      bool
}

func NewETSI014HTTPQuantumModule(addr, kmsId, slaveSAEID, masterSAEID string, tlsConfig config.TLSConfig, master bool) (*ETSI014HTTPQuantumModule, error) {
	parsedUrl, err := url.Parse(addr)
	if err != nil {
		return nil, err
	}

	restClientConf := &etsi14ClientGenerated.Configuration{
		Debug: true,
		Servers: etsi14ClientGenerated.ServerConfigurations{
			{
				URL:         parsedUrl.String(),
				Description: "QKD Module with ETSI14 implemented as API.",
			},
		},
		Scheme: parsedUrl.Scheme,
	}

	if tlsConfig.TLS {
		tlsConf, err := kmstls.GenerateTlsLibraryConfig(tlsConfig)
		if err != nil {
			return nil, fmt.Errorf("unable to generate TLS config: %w", err)
		}

		restClientConf.HTTPClient = &http.Client{
			Transport: &http.Transport{
				TLSClientConfig: tlsConf,
			},
		}
	}

	client, err := etsi14ClientImpl.NewClientImpl(restClientConf)
	if err != nil {
		return nil, err
	}

	return &ETSI014HTTPQuantumModule{
		id:          uuid.New(),
		kmsId:       kmsId,
		addr:        addr,
		keyStore:    store.NewKmsKeyStore(256),
		kmsClient:   nil,
		client:      client,
		SlaveSAEID:  slaveSAEID,
		MasterSAEID: masterSAEID,
		master:      master,
	}, nil
}

func (qm *ETSI014HTTPQuantumModule) ID() uuid.UUID {
	return qm.id
}

func (qm *ETSI014HTTPQuantumModule) Client() *etsi14ClientImpl.ClientImpl {
	return qm.client
}

func (qm *ETSI014HTTPQuantumModule) Initialize() error {
	// start polling
	if qm.master {
		go func() {
			ticker := time.NewTicker(2 * time.Second)
			defer ticker.Stop()

			// TODO: add context/channel to stop
			for range ticker.C {
				container, _, err := qm.client.GetKey(qm.SlaveSAEID)
				if err != nil {
					log.Error(err)
					break
				}

				keyIds := make([]string, len(container.GetKeys()))
				for i, keyItem := range container.GetKeys() {
					keyIds[i] = keyItem.GetKeyID()
				}

				_, err = qm.kmsClient.KeyIdNotification(context.Background(),
					&pbIC.KeyIdNotificationRequest{
						Timestamp: time.Now().Unix(),
						KmsId:     qm.kmsId,
						KeyIds:    keyIds,
					})
				if err != nil {
					log.Error(err)
					break
				}

				if err := store.AddETSIKeysToKeystore(qm.keyStore, container.GetKeys()); err != nil {
					log.Error(err)
				}
			}
		}()
	}
	return nil
}

func (qm *ETSI014HTTPQuantumModule) SetKmsPeerInformation(kmsClient *GRPCClient, kmsEventBus *event.EventBus, kmsTcpSocketStr string) error {
	qm.kmsClient = kmsClient
	return nil
}

func (qm *ETSI014HTTPQuantumModule) Address() string {
	return qm.addr
}

func (qm *ETSI014HTTPQuantumModule) KeyStore() *store.KmsKeyStore {
	return qm.keyStore
}

func (qm *ETSI014HTTPQuantumModule) SetKeyStore(newKeyStore *store.KmsKeyStore) {
	qm.keyStore = newKeyStore
}

func (qm *ETSI014HTTPQuantumModule) Sync() error {
	return nil
}

func (qm *ETSI014HTTPQuantumModule) KmsGrpcClient() *GRPCClient {
	return nil
}

func (qm *ETSI014HTTPQuantumModule) SetKmsGrpcClient(peer *GRPCClient) {
}
