package peers

import (
	"context"
	"encoding/base64"
	"fmt"
	"net"
	"time"

	pbIC "code.fbi.h-da.de/danet/quant/goKMS/api/gen/proto/go/kmsintercom"
	"code.fbi.h-da.de/danet/quant/goKMS/kms/crypto"
	"code.fbi.h-da.de/danet/quant/goKMS/kms/event"
	"code.fbi.h-da.de/danet/quant/goKMS/kms/util"
	"github.com/google/uuid"
	log "github.com/sirupsen/logrus"
)

// NOTE: currently not used, could be of usage later on
// type kmsPeerInfo interface {
// 	GetKmsPeerStatus() KmsPeerStatus
// 	GetKmsPeerId() uuid.UUID
// 	GetKmsPeerQkdiId() uint32
// 	KmsPeerKeyInit()
// }

type KmsPeerStatus int16

const (
	KmsPeerUp      KmsPeerStatus = iota // peer was seen up lately
	KmsPeerDown                         // peer not reachable anymore
	KmsPeerPending                      // peer was added and has to be successfully contacted once
	KmsPeerUnknown                      // not known, not initialized
)

type GRPCClient struct {
	pbIC.KmsTalkerClient
}

type KmsPeer struct {
	peerClient          *GRPCClient
	peerStatus          KmsPeerStatus
	peerKmsId           uuid.UUID
	servingQuantumModul QuantumModule
	tcpSocket           *net.TCPAddr // the IP address and TCP port (aka socket) of the kms peer
	TcpSocketStr        string       // string rep. of tcpSocket
	et                  crypto.CryptoAlgorithm
	// NOTE: currently not used, could be of usage later on
	// name                string    // the name of the kms peer
	quit     chan bool // cancel the peer goroutine
	eventBus *event.EventBus
}

// TODO: check intercomaddr -> remove?
func NewKmsPeer(peerKmsId string, quantummodule QuantumModule, tcpSocketStr string, client *GRPCClient, eventBus *event.EventBus) (*KmsPeer, error) {
	var peerKmsIdUUID uuid.UUID
	if peerKmsId == "" {
		peerKmsIdUUID = uuid.New()
	} else {
		id, err := uuid.Parse(peerKmsId)
		if err != nil {
			return nil, err
		}

		peerKmsIdUUID = id
	}

	tcpSocket, err := net.ResolveTCPAddr("tcp", tcpSocketStr)
	if err != nil {
		return nil, err
	}

	err = quantummodule.SetKmsPeerInformation(client, eventBus, tcpSocketStr)
	if err != nil {
		return nil, err
	}

	kmsPeer := &KmsPeer{
		// NOTE It could be a good idea to turn client and status into its own
		// struct (and additional information) e.g. a link
		// We need multiple peer clients!
		peerClient: client,
		// TODO: change this, only for demo purposes
		peerStatus: KmsPeerUp,
		peerKmsId:  peerKmsIdUUID,
		// NOTE a peer could have multiple quantum modules
		servingQuantumModul: quantummodule,
		tcpSocket:           tcpSocket,
		TcpSocketStr:        tcpSocketStr,
		et:                  crypto.NewAES(),
		quit:                make(chan bool),
		eventBus:            eventBus,
	}

	if err := quantummodule.Initialize(); err != nil {
		return nil, err
	}

	return kmsPeer, nil
}

func (kp *KmsPeer) Client() *GRPCClient {
	return kp.peerClient
}

func (kp *KmsPeer) EventBus() *event.EventBus {
	return kp.eventBus
}

func (kp *KmsPeer) Address() *net.TCPAddr {
	return kp.tcpSocket
}

func (kp *KmsPeer) QuantumModule() QuantumModule {
	return kp.servingQuantumModul
}

func (kp *KmsPeer) CryptoAlgo() crypto.CryptoAlgorithm {
	return kp.et
}

func (kp *KmsPeer) SyncBulkKeys() error {
	err := kp.servingQuantumModul.Sync()
	if err != nil {
		kp.peerStatus = KmsPeerUp
		return nil
	}
	return err
}

// TransportKeyNegotiation sends a request for a transport key negotiation to
// the peer KMS. The key that was agreed upon, is used to create a encrypted
// payload.
func (kp *KmsPeer) TransportKeyNegotiation() error {
	// ctx, cancel := context.WithTimeout(context.Background(), time.Second)
	// req, err := ph.peerClient.InterComTransportKeyNegotiation()
	return nil
}

func (kp *KmsPeer) SendInitialPayloadBasedOnGRPCClient(key *crypto.Key, pathId, processId uuid.UUID, kmsId string, remoteKMS *util.RemoteKMS) error {
	if kp.peerClient.KmsTalkerClient != nil {
		return kp.SendPayload(key, pathId, processId)
	}
	return fmt.Errorf("could not find a valid peer client")
}

func (kp *KmsPeer) SendPayload(payload *crypto.Key, pathId, processId uuid.UUID) error {
	// NOTE: It should be assumed that there are keys available if we try to
	// send.
	if kp.servingQuantumModul.KeyStore().Length() == 0 {
		log.Debug("Syncing bulk keys before sending the payload")
		err := kp.SyncBulkKeys()
		if err != nil {
			return err
		}
	}

	// select a key from key store
	key, err := kp.servingQuantumModul.KeyStore().GetKey()
	if err != nil {
		return err
	}

	// Start the negotiation process for a transport key
	ctx, cancel := context.WithTimeout(context.Background(), time.Second)
	defer cancel()
	_, err = kp.peerClient.InterComTransportKeyNegotiation(ctx, &pbIC.InterComTransportKeyNegotiationRequest{
		Timestamp: time.Now().Unix(),
		PathID:    pathId.String(),
		KeyToUse:  key.KeyID.String(),
	})
	if err != nil {
		return err
	}

	kp.servingQuantumModul.KeyStore().DeleteKey(key.KeyID)

	nonce, encryptedPayload, err := kp.et.Encrypt(payload.Key, key.Key)
	if err != nil {
		return err
	}

	encryptedPayloadAsString := base64.StdEncoding.EncodeToString(encryptedPayload)
	nonceAsString := base64.StdEncoding.EncodeToString(nonce)

	log.Infof("Sent encrypted Payload: %s with nonce: %s", encryptedPayloadAsString, nonceAsString)

	ctx2, cancel2 := context.WithTimeout(context.Background(), time.Second)
	defer cancel2()
	_, err = kp.peerClient.KeyForwarding(ctx2, &pbIC.KeyForwardingRequest{
		Timestamp: time.Now().Unix(),
		PathId:    pathId.String(),
		ProcessId: processId.String(),
		Key: &pbIC.Key{
			Id:    payload.ID.String(),
			Nonce: nonceAsString,
			Key:   encryptedPayloadAsString,
		},
	})
	if err != nil {
		return err
	}

	return nil
}

func (kp *KmsPeer) Status() KmsPeerStatus {
	return kp.peerStatus
}

func (kp *KmsPeer) SetStatus(updateStatus KmsPeerStatus) {
	kp.peerStatus = updateStatus
}

func (kp *KmsPeer) GetKmsPeerId() uuid.UUID {
	return kp.peerKmsId
}
