package store

import (
	"encoding/base64"
	"fmt"
	"sync"

	etsi14 "code.fbi.h-da.de/danet/quant/etsi014/go/rest/etsi/client"
	"github.com/google/uuid"
	log "github.com/sirupsen/logrus"
)

type Status int

const (
	RESERVED = iota
	AVAILABLE
	UNKNOWN
	USED
)

// holds a single ready to bit key, length can be configured.
type KmsKSElement struct {
	KeyID  uuid.UUID
	Key    []byte // a key
	Status Status
}

type KmsKeyStore struct {
	keyStoreMutex sync.Mutex
	keyStore      map[uuid.UUID]*KmsKSElement
	KeySingleSize uint // the size of a single key, given as unit of bits
}

func NewKmsKeyStore(desiredkeySingleSizeLength uint) *KmsKeyStore {
	return &KmsKeyStore{
		// NOTE: could be useful to use something like a ordered map here
		keyStore:      make(map[uuid.UUID]*KmsKSElement),
		KeySingleSize: desiredkeySingleSizeLength,
	}
}

func (ks *KmsKeyStore) Length() int {
	ks.keyStoreMutex.Lock()
	defer ks.keyStoreMutex.Unlock()

	return len(ks.keyStore)
}

func (ks *KmsKeyStore) AddKey(keyId uuid.UUID, keyToadd []byte) {
	ks.keyStoreMutex.Lock()
	defer ks.keyStoreMutex.Unlock()

	// test for collisions
	if _, notThere := ks.keyStore[keyId]; notThere {
		log.Errorf("Whop: addKey collisions of key id %s", keyId)
		return
	}

	newKeyElement := &KmsKSElement{
		KeyID:  keyId,
		Key:    keyToadd,
		Status: AVAILABLE,
	}
	// ok to add
	ks.keyStore[newKeyElement.KeyID] = newKeyElement
}

func (ks *KmsKeyStore) GetKey() (*KmsKSElement, error) {
	ks.keyStoreMutex.Lock()
	defer ks.keyStoreMutex.Unlock()

	for _, key := range ks.keyStore {
		if key.Status == AVAILABLE {
			// change status of key to reserved
			key.Status = RESERVED
			return key, nil
		}
	}
	return nil, fmt.Errorf("could not find a key to use")
}

func (ks *KmsKeyStore) GetKeyWithID(id uuid.UUID) (*KmsKSElement, error) {
	ks.keyStoreMutex.Lock()
	defer ks.keyStoreMutex.Unlock()

	key, ok := ks.keyStore[id]
	if !ok {
		return nil, fmt.Errorf("Could not find a key for id: %v", id)
	}
	if key.Status == AVAILABLE {
		// change status of key to reserved
		key.Status = RESERVED
		return key, nil
	}
	return nil, fmt.Errorf("Key with index: %v is already reserved.", id)
}

func (ks *KmsKeyStore) DeleteKey(keyId uuid.UUID) {
	ks.keyStoreMutex.Lock()
	defer ks.keyStoreMutex.Unlock()
	delete(ks.keyStore, keyId)
}

func AddETSIKeysToKeystore(keyStore *KmsKeyStore, keyContainer []etsi14.KeyContainerKeysInner) error {
	for _, keyItem := range keyContainer {
		// decode base64 encoded key string
		key, err := base64.StdEncoding.DecodeString(keyItem.GetKey())
		if err != nil {
			return err
		}
		keyId, err := uuid.Parse(keyItem.GetKeyID())
		if err != nil {
			return err
		}
		// add to keystore
		keyStore.AddKey(keyId, key)
	}
	return nil
}
